'use strict';

// const assert = require('assert');
const { GraphQLNonNull, GraphQLList, getNamedType } = require('graphql');
const gql = require('graphql-tag');
const { introspectSchema } = require('graphql-tools');
const { ApolloClient } = require('apollo-client');
const { ApolloLink } = require('apollo-link');
const { createHttpLink } = require('apollo-link-http');
// const {onError} = require('apollo-link-error');
// const {setContext} = require('apollo-link-context');
const { RetryLink } = require('apollo-link-retry');
const { createPersistedQueryLink } = require('apollo-link-persisted-queries');
const { InMemoryCache } = require('apollo-cache-inmemory');
const fetch = require('node-fetch');

class Client {
  constructor(uri, httpOptions = {}) {
    this.uri = uri;
    this.links = {
      retry: new RetryLink({
        delay: {
          initial: 300,
          max: 5 * 60 * 1000,
          jitter: true,
        },
        attempts: {
          max: 20,
          retryIf: error => !!error,
        },
      }),
      persistedQuery: createPersistedQueryLink(),
      http: createHttpLink({ uri, fetch, ...httpOptions }),
    };
    this.apolloClient = new ApolloClient({
      link: ApolloLink.from([ 'retry', 'persistedQuery', 'http' ].map(l => this.links[l])),
      cache: new InMemoryCache(),
      queryDeduplication: false,
      defaultOptions: { query: { fetchPolicy: 'no-cache' } },
    });
  }
  async connect() {
    this.schema = await introspectSchema(this.links.http);
    for (const field of Object.values(this.schema.getQueryType().getFields())) {
      const queryArgDeclare = field.args.map(a => `$${a.name}: ${makeFieldTypeExpr(a.type)}`).join(', ');
      const queryArgs = field.args.map(a => `${a.name}: $${a.name}`).join(', ');
      let returnTypeFields = getNamedType(field.type).getFields();

      // const hasResults = returnTypeFields.results;
      // if (hasResults) {
      //   returnTypeFields = getNamedType(returnTypeFields.results.type).getFields();
      // }
      let defaultProjections = { [returnTypeFields.updatedAt ? 'updatedAt' : Object.keys(returnTypeFields)[0]]: true };
      // if (hasResults) {
      //   defaultProjections = { results: defaultProjections };
      // }
      this[`query${field.name}`] = async (variables, projections) => {
        if (!projections || Object.keys(projections).length <= 0) projections = defaultProjections;
        try {
          const resp = await this.apolloClient.query({
            query: gql`query (${queryArgDeclare}) { ${field.name} (${queryArgs}) ${makeReturnExpr(projections)} }`,
            variables,
          });
          return resp.data[field.name];
        } catch (e) {
          if (e.networkError && (!e.networkError.statusCode || e.networkError.statusCode >= 500)) {
            this.logger.fail('_failed_api_server_error', e);
          } else {
            throw e;
          }
        }
      };
    }
    for (const field of Object.values(this.schema.getMutationType().getFields())) {
      const queryArgDeclare = field.args.map(a => `$${a.name}: ${makeFieldTypeExpr(a.type)}`).join(', ');
      const queryArgs = field.args.map(a => `${a.name}: $${a.name}`).join(', ');
      let returnTypeFields = getNamedType(field.type).getFields();
      const hasResults = returnTypeFields.results;
      // if (hasResults) {
      //   returnTypeFields = getNamedType(returnTypeFields.results.type).getFields();
      // }
      let defaultProjections = { [returnTypeFields.updatedAt ? 'updatedAt' : Object.keys(returnTypeFields)[0]]: true };
      // if (hasResults) {
      //   defaultProjections = { results: defaultProjections };
      // }
      this[`${field.name[0].toLowerCase()}${field.name.slice(1)}`] = async (variables, projections) => {
        if (!projections || Object.keys(projections).length <= 0) projections = defaultProjections;
        try {
          const resp = await this.apolloClient.mutate({
            mutation: gql`mutation (${queryArgDeclare}) { ${field.name} (${queryArgs}) ${makeReturnExpr(projections)} }`,
            variables,
          });
          return resp.data[field.name];
        } catch (e) {
          if (e.networkError && (!e.networkError.statusCode || e.networkError.statusCode >= 500)) {
            this.logger.fail('_failed_api_server_error', e);
          } else {
            throw e;
          }
        }
      };
    }
  }
}

function makeFieldTypeExpr(type) {
  if (type instanceof GraphQLNonNull) {
    return `${makeFieldTypeExpr(type.ofType)}!`;
  }
  if (type instanceof GraphQLList) {
    return `[${makeFieldTypeExpr(type.ofType)}]`;
  }
  return type.name;
}

/* eslint array-callback-return: ["error", { allowImplicit: true }]*/
function makeReturnExpr(projections) {
  return `{${Object.entries(projections)
    .map(([ field, show ]) => {
      if (typeof show === 'object') {
        const expr = makeReturnExpr(show);
        if (expr === '{}') {
          return;
        }
        return `${field} ${expr}`;
      }
      if (show) {
        return field;
      }
      return;
    })
    .filter(v => v)
    .join(', ')}}`;
}

module.exports = async app => {
  const { API_HOST, API_PORT, API_PATH, httpOptions } = app.config.graphqlQuery;
  const uri = `http://${API_HOST}:${API_PORT}/${API_PATH}`;
  const client = new Client(uri, httpOptions);
  await client.connect();
  app.graphqlQuery = client;
};
